import type FoliageTransactionBlock from './FoliageTransactionBlock';
import type Foliage from './Foliage';

type Block = {
  foliage_transaction_block: FoliageTransactionBlock;
  foliage: Foliage;
};

export default Block;
