export const STANDARD_WALLET = 0;
export const RATE_LIMITED = 1;
export const ATOMIC_SWAP = 2;
export const AUTHORIZED_PAYEE = 3;
export const MULTI_SIG = 4;
export const CUSTODY = 5;
export const COLOURED_COIN = 6;
export const RECOVERABLE = 7;
export const DISTRIBUTED_ID = 8;
