import { createTeleporter } from 'react-teleporter';

const PoolHeaderTeleporter = createTeleporter();

export const PoolHeaderSource = PoolHeaderTeleporter.Source;

export const PoolHeaderTarget = PoolHeaderTeleporter.Target;
