export default {
  palette: {
    primary: {
      main: '#3AAC59',
      contrastText: '#ffffff',
    },
    secondary: {
      main: '#000000',
      contrastText: '#ffffff',
    },
    danger: {
      main: '#dc3545',
      contrastText: '#ffffff',
    },
  },
  drawer: {
    width: '100px',
  },
};
