import { createGlobalStyle } from 'styled-components';
import '@fontsource/roboto/700.css';
import '@fontsource/roboto/500.css';
import '@fontsource/roboto/400.css';
import '@fontsource/roboto/300.css';

export default createGlobalStyle`
  body {
    font-family: "Roboto";
  }
`;
